#include <stdlib.h>
#include "lph.h"

void port_init(void)
{
  uint8_t tmp;

  PORTB |= _BV(LCD_MOSI);
  DDRB |= _BV(LCD_MOSI);
  
  PORTB &= ~_BV(LCD_CS);
  DDRB |= _BV(LCD_CS);
  
  PORTB &= ~_BV(LCD_SS);
  DDRB |= _BV(LCD_SS);
  
  PORTB &= ~_BV(LCD_SCK);
  DDRB |= _BV(LCD_SCK);
  
  PORTB |= _BV(LCD_MISO);

  // setup SPI Interface
  SPCR |= _BV(MSTR) | _BV(SPE);
  SPSR |= _BV(SPI2X);  // double speed bit

  tmp=SPSR;
}

void lcd_write(uint8_t dat)
{
  SPDR= dat;
  while (!(SPSR & _BV(SPIF)));
}

void lcd_cspulse(void)
{
  PORTB |= _BV(LCD_CS);
  asm volatile("nop");
  PORTB &= ~_BV(LCD_CS);
}


void lcd_comtype(uint8_t dat1)
{
  lcd_write(0x74);  // indicates command type
  lcd_write(0x00);
  lcd_write(dat1);
  lcd_cspulse();
}

void lcd_comdat(uint8_t dat1, uint8_t dat2)
{
  lcd_write(0x76);  // indicates command data
  lcd_write(dat1);
  lcd_write(dat2);
  lcd_cspulse();
}

void lcd_write16(uint16_t dat)
{
  lcd_write(dat>>8);
  lcd_write(dat);
}

void lcd_init(void)
{
	PORTB &= ~_BV(LCD_CS);
	
	lcd_comtype(0x00);
	lcd_comdat(0x00, 0x01);
	_delay_ms(10);
	lcd_comtype(0x01);
	lcd_comdat(0x00, 0x15);
	lcd_comtype(0x02);
	lcd_comdat(0x04, 0x00);
	lcd_comtype(0x03);
	lcd_comdat(0x00, 0x0C);
	lcd_comtype(0x04);
	lcd_comdat(0x00, 0x00);
	lcd_comtype(0x05);
	lcd_comdat(0x00, 0x38);
	lcd_comtype(0x06);
	lcd_comdat(0x00, 0x00);
	lcd_comtype(0x07);
	lcd_comdat(0x00, 0x37);
	lcd_comtype(0x0B);
	lcd_comdat(0x00, 0x00);
	lcd_comtype(0x0C);
	lcd_comdat(0x00, 0x01);
	lcd_comtype(0x0D);
	lcd_comdat(0x06, 0x16);
	lcd_comtype(0x0E);
	lcd_comdat(0x2D, 0x1F);
	lcd_comtype(0x0F);
	lcd_comdat(0x00, 0x00);
	lcd_comtype(0x11);
	lcd_comdat(0x00, 0x00);
	lcd_comtype(0x14);
	lcd_comdat(0xAF, 0x00);
	lcd_comtype(0x15);
	lcd_comdat(0x00, 0x00);
	lcd_comtype(0x16);
	lcd_comdat(0x83, 0x00);
	lcd_comtype(0x17);
	lcd_comdat(0xAF, 0x00);
	lcd_comtype(0x20);
	lcd_comdat(0x00, 0x00);
	lcd_comtype(0x21);
	lcd_comdat(0x00, 0x00);
	lcd_comtype(0x30);
	lcd_comdat(0x00, 0x00);
	lcd_comtype(0x31);
	lcd_comdat(0x04, 0x00);
	lcd_comtype(0x32);
	lcd_comdat(0x02, 0x07);
	lcd_comtype(0x33);
	lcd_comdat(0x07, 0x00);
	lcd_comtype(0x34);
	lcd_comdat(0x00, 0x05);
	lcd_comtype(0x35);
	lcd_comdat(0x07, 0x03);
	lcd_comtype(0x36);
	lcd_comdat(0x07, 0x07);
	lcd_comtype(0x37);
	lcd_comdat(0x00, 0x07);
	lcd_comtype(0x38);
	lcd_comdat(0x18, 0x19);
	lcd_comtype(0x39);
	lcd_comdat(0x1A, 0x1B);
	lcd_comtype(0x3A);
	lcd_comdat(0x12, 0x00);
	lcd_comtype(0x3B);
	lcd_comdat(0x00, 0x09);
	lcd_comtype(0x3C);
	lcd_comdat(0x22, 0x24);
	lcd_comtype(0x3D);
	lcd_comdat(0x26, 0x28);
	lcd_comtype(0x3E);
	lcd_comdat(0x2B, 0x2E);
	lcd_comtype(0x3F);
	lcd_comdat(0x31, 0x34);
	
	PORTB |= _BV(LCD_CS);
}

void fill_rect(uint8_t start_x, uint8_t start_y, uint8_t end_x, uint8_t end_y, uint16_t color)
{
	uint16_t i, count;
	
	
	count = (end_x - start_x) * (end_y - start_y) * 2;

	PORTB &= ~_BV(LCD_CS);  // select display
	
	lcd_comtype(0x16);
	lcd_comdat(end_y, start_y);
	lcd_comtype(0x17);
	lcd_comdat(end_x, start_x);
	lcd_comtype(0x21);
	lcd_comdat(0x00, 0x00);
	lcd_comtype(0x22);
	lcd_write(0x76);  

	// start data transmission  
	for (i=0; i<count; i++)
	{
		lcd_write16(color);
	}
	PORTB |= _BV(LCD_CS);
	PORTB &= ~_BV(LCD_CS);
	lcd_comtype(0x16);
	lcd_comdat(0xAF, 0x00);
	lcd_comtype(0x17);
	lcd_comdat(0x83, 0x00);
	
	PORTB |= _BV(LCD_CS);  // deselect display
}

void set_pixel(uint8_t pos_x, uint8_t pos_y, uint16_t color)
{
	PORTB &= ~_BV(LCD_CS);
	
	lcd_comtype(0x21);
	lcd_comdat(pos_x, pos_y);
	lcd_comtype(0x22);
	lcd_write(0x76);
	lcd_write16(color);
	
	PORTB |= _BV(LCD_CS);
}

void set_window(uint8_t start_x, uint8_t start_y, uint8_t end_x, uint8_t end_y)
{
	PORTB &= ~_BV(LCD_CS);
	
	lcd_comtype(0x16);
	lcd_comdat(end_y, start_y);
	lcd_comtype(0x17);
	lcd_comdat(end_x, start_x);
	
	PORTB |= _BV(LCD_CS);
}

	/* line- and circle-function from a KS0108-library by F. Thiele */

void draw_line( uint8_t x1, uint8_t y1, uint8_t x2, uint8_t y2, uint16_t color) 
 { uint8_t length, xTmp, yTmp, i, y, yAlt;
   int16_t m;

   if(x1 == x2) 
    { // vertical line
      // x1|y1 must be the upper point
      if(y1 > y2) 
       { xTmp = x1;
         yTmp = y1;
         x1 = x2;
         y1 = y2;
         x2 = xTmp;
         y2 = yTmp;
       }
      length = y2-y1;
      for(i=0; i<=length; i++) 
         set_pixel(x1, y1+i, color);
    } 
   else if(y1 == y2) 
    { // horizontal line
      // x1|y1 must be the left point
      if(x1 > x2) 
       { xTmp = x1;
         yTmp = y1;
         x1 = x2;
         y1 = y2;
         x2 = xTmp;
         y2 = yTmp;
       }

      length = x2-x1;
      for(i=0; i<=length; i++) 
         set_pixel(x1+i, y1, color);
       
    } 
   else 
    { // x1 must be smaller than x2
      if(x1 > x2) 
       { xTmp = x1;
         yTmp = y1;
         x1 = x2;
         y1 = y2;
         x2 = xTmp;
         y2 = yTmp;
       }
		
      if((y2-y1) >= (x2-x1) || (y1-y2) >= (x2-x1)) 
       { // angle larger or equal 45
         length = x2-x1;								// not really the length :)
         m = ((y2-y1)*200)/length;
         yAlt = y1;
         for(i=0; i<=length; i++) 
          { y = ((m*i)/200)+y1;
            if((m*i)%200 >= 100)
               y++;
            else if((m*i)%200 <= -100)
               y--;
				
            draw_line(x1+i, yAlt, x1+i, y, color); /* wuff wuff recurs. */
            if(length<=(y2-y1) && y1 < y2)
               yAlt = y+1;
            else if(length<=(y1-y2) && y1 > y2)
               yAlt = y-1;
            else
               yAlt = y;
          }
       } 
      else
       { // angle smaller 45
         // y1 must be smaller than y2
         if(y1 > y2)
          { xTmp = x1;
            yTmp = y1;
            x1 = x2;
            y1 = y2;
            x2 = xTmp;
            y2 = yTmp;
          }
         length = y2-y1;
         m = ((x2-x1)*200)/length;
         yAlt = x1;
         for(i=0; i<=length; i++)
          { y = ((m*i)/200)+x1;

            if((m*i)%200 >= 100)
               y++;
            else if((m*i)%200 <= -100)
               y--;

            draw_line(yAlt, y1+i, y, y1+i, color); /* wuff */
            if(length<=(x2-x1) && x1 < x2)
               yAlt = y+1;
            else if(length<=(x1-x2) && x1 > x2)
               yAlt = y-1;
            else
               yAlt = y;
          }
       }
    }
 }
	
void draw_circle(const uint8_t xCenter, const uint8_t yCenter, const uint8_t radius, const uint8_t color) 
 { int16_t tSwitch, y, x = 0;
   uint8_t d;
   
   d = yCenter - xCenter;
   y = radius;
   tSwitch = 3 - 2 * radius;

   while (x <= y) 
    { 
		set_pixel(xCenter + x, yCenter + y, color);
		set_pixel(xCenter + x, yCenter - y, color);
	
		set_pixel(xCenter - x, yCenter + y, color);
		set_pixel(xCenter - x, yCenter - y, color);
	
		set_pixel(yCenter + y - d, yCenter + x, color);
		set_pixel(yCenter + y - d, yCenter - x, color);
   
		set_pixel(yCenter - y - d, yCenter + x, color);
		set_pixel(yCenter - y - d, yCenter - x, color);

		if (tSwitch < 0)
		{
			tSwitch += (4 * x + 6);
		}
		
		else 
		{
			tSwitch += (4 * (x - y) + 10);
			y--;
		}
		x++;
    }
 }

/*
 Writes a glyph("letter") to the display at location x,y
 (adapted function from the MJK-code)
 Arguments are:
    column    - x corrdinate of the left part of glyph          
    row       - y coordinate of the top part of glyph       
    width     - size in pixels of the width of the glyph    
    height    - size in pixels of the height of the glyph   
    glyph     - an unsigned char pointer to the glyph pixels 
                to write assumed to be of length "width"
*/



void lcd_glyph(uint8_t start_left, uint8_t start_top, uint8_t width, uint8_t height, uint8_t *glyph_ptr, uint8_t store_width)
{
	uint8_t mask;
	uint8_t char_mask;
//	uint8_t height_counter;
//	uint8_t width_counter;
//	uint8_t need_bytes;
	uint8_t counter;
//	uint8_t counter_x;
	uint8_t counter_y;
//	uint8_t use_bits;
	
	for(counter_y = 0; counter_y < height; counter_y++)
	{
	
		mask = *glyph_ptr;
		
		char_mask = 128;
		for(counter = 0; counter <= width; counter++)
		{
			if(pgm_read_byte(glyph_ptr) & char_mask)
			{
				set_pixel(start_left + counter, start_top + counter_y, 0x470d);
			}
			else
			{
				set_pixel(start_left + counter, start_top + counter_y, 0xffff);
			}
			char_mask /= 2;
		}
		glyph_ptr += store_width;
	}
	

}

/*
 Prints the given string at location x,y in the specified font.
 Prints each character given via calls to lcd_glyph. The entry string
 is null terminated. (adapted function from the MJK-code)
 Arguments are:
	left       coordinate of left start of string.
	top        coordinate of top of string.
	font       font number to use for display (see fonts.h)
	str        text string to display (null-terminated)
*/
static void lcd_text_intern(uint8_t left, uint8_t top, uint8_t font, const char *str, uint8_t inprogmem)
 { uint8_t x = left;
   uint8_t glyph;
   uint8_t width;
   uint8_t height, defaultheight;
   uint8_t store_width;
   uint8_t *glyph_ptr;
   uint8_t *width_table_ptr;
   uint8_t *glyph_table_ptr;
   uint8_t glyph_beg, glyph_end;
   uint8_t fixedwidth;

   defaultheight = pgm_read_byte ( &(fonts[font].glyph_height) );
   store_width = pgm_read_byte ( &(fonts[font].store_width) );
   width_table_ptr = (uint8_t*) pgm_read_word( &(fonts[font].width_table) );
   glyph_table_ptr = (uint8_t*) pgm_read_word( &(fonts[font].glyph_table) );
   glyph_beg  = pgm_read_byte( &(fonts[font].glyph_beg) );
   glyph_end  = pgm_read_byte( &(fonts[font].glyph_end) );
   fixedwidth = pgm_read_byte( &(fonts[font].fixed_width) );

   if (inprogmem) 
      glyph = pgm_read_byte(str);
    
   else 
      glyph = (uint8_t)*str;
	
   while(glyph != 0x00) // while(*str != 0x00)
    { /* check to make sure the symbol is a legal one */
      /* if not then just replace it with the default character */
      if((glyph < glyph_beg) || (glyph > glyph_end))
         glyph = pgm_read_byte( &(fonts[font].glyph_def) ) ;

      /* make zero based index into the font data arrays */
      glyph -= glyph_beg;
      if(fixedwidth == 0)
         // width=fonts[font].width_table[glyph];	/* get the variable width instead */
         width=pgm_read_byte(width_table_ptr+glyph);
      else 
         width = fixedwidth;
		
      height = defaultheight;
      //glyph_ptr = fonts[font].glyph_table + ((unsigned int)glyph * (unsigned int)store_width * (unsigned int)height);
      glyph_ptr = glyph_table_ptr + ((unsigned int)glyph * (unsigned int)store_width * (unsigned int)height) ;

      /* range check / limit things here */
      if(x > SCRN_RIGHT)
         x = SCRN_RIGHT;
       
      if((x + width) > SCRN_RIGHT+1)
         width = SCRN_RIGHT - x + 1;
       
      if(top > SCRN_BOTTOM)
         top = SCRN_BOTTOM;
       
      if((top + height) > SCRN_BOTTOM+1)
         height = SCRN_BOTTOM - top + 1;
       
      lcd_glyph(x,top,width,height,glyph_ptr,store_width);  /* plug symbol into buffer */

      x += width;		/* move right for next character */
      str++;			/* point to next character in string */
      if (inprogmem) 
         glyph = pgm_read_byte(str);
      else 
         glyph = (uint8_t)*str;
       
    }
 }

/* draw string from RAM */
void lcd_text(uint8_t left, uint8_t top, uint8_t font, const char *str)
 { lcd_text_intern(left, top, font, str, 0);
 }

/* draw string from PROGMEM */
void lcd_text_p(uint8_t left, uint8_t top, uint8_t font, const char *str)
 { lcd_text_intern(left, top, font, str, 1);
 }

